#!/usr/bin/env python3

import argparse
import sys
import logging

from configure_functions import UnknownAuthType, TokenGenerationFailed, parse_configuration_file, SonarqubeApi


def requires_restart(args):
    if args.install or args.update_plugins or args.restart:
        return True
    return False


parser = argparse.ArgumentParser(description='Configure sonarqube instance',
                                 usage="""
# auth by token
  $ python configure.py https://sonarqube-dev.app.corp gfdsjgkldfjgkljdflkgdfg --plugins xanitizer other --list
# basic auth
  $ python configure.py https://sonarqube-dev.app.corp admin --password admin --auth-type basic --generate-token test""",
                                 )
parser.add_argument('host', help='host name of sonarqube server')
parser.add_argument('login', help='user login or user authentication key')
parser.add_argument('--password', help='user authentication key')
parser.add_argument('--auth-type', help='type of authentication', default='key', choices=['basic', 'key'])
parser.add_argument('--install', help='plugins to install', nargs='+')
parser.add_argument('--update-plugins', help='update installed plugins', choices=['all', 'outdated'])
parser.add_argument('--list', help='list installed plugins', action='store_true')
parser.add_argument('--verbose', help='makes the output more verbose', action='store_true')
parser.add_argument('--logger-level', help='logger level to control output verbosity', default='INFO', choices=logging._nameToLevel.keys())
parser.add_argument('--restart', help='restart arbitrarly', action='store_true')
parser.add_argument('--force-restart', help='forces restart', action='store_true')
parser.add_argument('--update-settings', help='update settings from a properties file')
parser.add_argument('--create-azure-user', help='creates the azure user, grants it admin privileges and creates a token', metavar='KEYVAULT_NAME')
parser.add_argument('--max-retries',
                    help='number of retries to contact the server before timing out (each try is separated by 1s)',
                    type=int, default=360)
parser.add_argument('--generate-token',
                    help='generate, output an authentication token for the user and leaves immediatly')
parser.add_argument('--bootstrap-security-groups',
                    help='generate standard groups, users stereotypes and permission templates', action='store_true')
parser.add_argument('--init-project',
                    help='Creates the project and the project dev and admins groups', nargs=2, metavar=('project_key', 'project_name'))
try:
    args = parser.parse_args()

    api = SonarqubeApi(
        args.host,
        args.login,
        password=args.password,
        auth_type=args.auth_type,
        verbose=args.verbose,
        max_retries=args.max_retries)

    SG_TOKEN = "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"
    CONNECTION_STRING = "Server=10.170.98.35;Port=5432;Database=GeomApp;User ID=ppcsUser;Password=pPcsP@1234;"

    api.set_logger_level(args.logger_level)

    if not api.is_ready():
        sys.exit("Sonarqube Server is not ready to accept API calls")

    if not api.is_auth_valid():
        sys.exit(1)

    if args.generate_token:
        print(api.generate_token(args.generate_token))
        sys.exit(0)

    # list installed plugins
    if args.list:
        if args.verbose:
            plugins_descriptions = [
                '- [' + plugin.get('key') + '] : ' + plugin.get('name') + ' - ' + plugin.get('description')
                for plugin in api.plugins_installed()]
            print("Available plugins : \n" + "\n".join(plugins_descriptions))
        else:
            plugins_descriptions = [plugin.get('key') for plugin in api.plugins_installed()]
            print("Available plugins : " + " ".join(plugins_descriptions))

    # install plugins
    if args.install:
        print('Installing plugins...')
        [api.install_plugin(plugin) for plugin in args.install]

    # update settings from properties file
    if args.update_settings:
        print('Updating settings from file : ' + args.update_settings)
        settings = parse_configuration_file(args.update_settings)
        for key in settings.keys():
            api.update_setting(key, settings[key])

    if args.create_azure_user:
        api.create_azure_user(args.create_azure_user)

    # Update installed plugins
    if args.update_plugins == 'all':
        api.update_installed_plugins(update_all_plugins=True)
    elif args.update_plugins == 'outdated':
        api.update_installed_plugins()

    # Initialize Permissions and groups
    if args.bootstrap_security_groups:
        api.bootstrap_security_groups()

    # Initialize Permissions and groups
    if args.init_project:
        api.bootstrap_project_security_groups(args.init_project)

    # Restarts server if necessary
    if requires_restart(args):
        api.restart_server(args.force_restart)

except (TokenGenerationFailed, UnknownAuthType) as err:
    print(err.args[0])
