import requests
from requests.auth import HTTPBasicAuth
from azure.identity import EnvironmentCredential
from azure.keyvault.secrets import SecretClient
import urllib3
import logging
import colorlog
import time
import sys


class UnknownAuthType(ValueError):
    pass


class TokenGenerationFailed(ValueError):
    pass


def parse_configuration_file(path):
    """Reads a property configuration file and returns a dictionary : key -> value"""

    config_file = open(path, "r")
    config_lines = config_file.readlines()
    config_file.close()

    # strip comments and line breaks
    config_lines = [line for line in config_lines if not line.strip().startswith('#')]
    config_lines = [line.strip('\n') for line in config_lines]
    settings = dict()

    # separate key from value
    for config_line in config_lines:
        separator_index = config_line.find('=')
        key = config_line[:separator_index]
        value = config_line[(separator_index + 1):]
        settings[key] = value
    return settings


class SonarqubeApi:
    sonarqube_host = ""
    login = ""
    password = None
    auth_type = 'key'
    max_retries = 360
    verbose = False
    verify_certificate = False

    def __init__(
            self,
            sonarqube_host,
            login,
            password=None,
            auth_type='key',
            max_retries=360,
            verbose=False):

        self.sonarqube_host = sonarqube_host
        self.login = login
        self.max_retries = max_retries
        self.verbose = verbose
        self.password = password
        self.auth_type = auth_type

        handler = colorlog.StreamHandler()
        handler.setFormatter(colorlog.ColoredFormatter(
            '%(fg_yellow)s%(asctime)s%(reset)s %(log_color)s%(levelname)-8s%(reset)s [%(fg_cyan)s%(name)s%(reset)s.%(fg_blue)s%(funcName)s%(reset)s] %(message)s'))

        self.logger = colorlog.getLogger(self.__class__.__name__)

        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)-8s [%(name)s] %(message)s',
            datefmt="%Y-%m-%d %H:%M:%S",
            handlers=[handler]
        )

        if not verbose:
            # Disable Unverified HTTPS request warnings
            urllib3.disable_warnings()

    def set_logger_level(self, logger_level_to_use):
        self.logger.setLevel(logger_level_to_use)

    def _set_verify_certificate(self, verify_certificate:bool):
        self.verify_certificate = verify_certificate

    def auth(self):
        """REST auth depending on auth type"""

        if self.auth_type == 'key':
            return HTTPBasicAuth(self.login, '')
        if self.auth_type == 'basic':
            return HTTPBasicAuth(self.login, self.password)
        raise UnknownAuthType('Unknown auth type : "' + self.auth_type + '"')

    def __get(self, api_url: str, api_params: dict):
        self.logger.debug(
            f'__get(api_url={api_url}, api_params={api_params})')
        res = requests.get(
            self.sonarqube_host + api_url,
            auth=self.auth(),
            params=api_params,
            verify=self.verify_certificate
        )

        res.raise_for_status()

        return res

    def __post(self, api_url: str, api_params: dict):
        self.logger.debug(f'__post(api_url={api_url}, api_params={api_params})')
        res = requests.post(
            self.sonarqube_host + api_url,
            auth=self.auth(),
            params=api_params,
            verify=self.verify_certificate
        )

        res.raise_for_status()

        return res

    def is_ready(self):
        """Check if the server is ready to accept calls until timeout"""

        retries = 1
        server_is_ready = False
        while retries < self.max_retries and not server_is_ready:
            retries += 1

            if self.verbose:
                self.logger.debug('Checking server status. Iteration ' + str(retries))

            response = requests.get(
                self.sonarqube_host + "/api/system/status",
                auth=self.auth(),
                verify=self.verify_certificate)
            server_is_ready = response.ok and response.json().get('status') == 'UP'

            if not server_is_ready:
                time.sleep(1)

        return server_is_ready

    def is_auth_valid(self) -> bool:
        """Calls the API to check credentials"""
        auth_valid = self.__get(
            api_url='/api/authentication/validate',
            api_params={}
        )

        if auth_valid.ok:
            auth_valid_json = auth_valid.json()
            if 'valid' in auth_valid_json and auth_valid_json['valid']:
                self.logger.info('Authentication successful.')
                return True

        self.logger.error('Authentication failed.')
        return False

    def generate_token(self, token_name):
        """Generate an authentication token for the user"""

        if not self.auth_type == 'basic':
            raise TokenGenerationFailed('Token generation requires to be run in basic auth mode')

        result = requests.post(
            self.sonarqube_host + "/api/user_tokens/generate",
            auth=self.auth(),
            verify=self.verify_certificate,
            params={"login": self.login, "name": token_name})

        if not result.ok:
            raise TokenGenerationFailed(self.get_error_message(result))

        return result.json().get('token')

    def plugins_available_to_install(self):
        """A list of plugins not yet installed"""

        available_plugins_response = requests.get(
            self.sonarqube_host + "/api/plugins/available",
            auth=self.auth(),
            verify=self.verify_certificate)
        available_plugins_json = available_plugins_response.json().get('plugins')

        return available_plugins_json

    def plugins_to_update(self):
        """The list of plugins to update"""
        plugins_to_update_response = self.__get(
            api_url="/api/plugins/updates",
            api_params={}
        )

        if not plugins_to_update_response.ok:
            self.print_error_message(plugins_to_update_response)
            return {}

        return plugins_to_update_response.json().get('plugins')

    def plugins_installed(self):
        """A list of installed plugins"""

        available_plugins_response = self.__get(
            api_url="/api/plugins/installed",
            api_params={}
        )

        if not available_plugins_response.ok:
            self.print_error_message(available_plugins_response)
            return {}

        return available_plugins_response.json().get('plugins')

    def install_plugin(self, plugin):
        """Installs a plugin"""

        self.logger.info('Installing : ' + plugin + '...')
        result = requests.post(
            self.sonarqube_host + "/api/plugins/install",
            auth=self.auth(),
            verify=self.verify_certificate,
            params={"key": plugin})

        if not result.ok:
            self.print_error_message(result)

    def restart_server(self, force=False):
        if force or self.has_pended_plugins():
            self.logger.info('Restarting...')
            result = requests.post(
                self.sonarqube_host + "/api/system/restart",
                auth=self.auth(),
                verify=self.verify_certificate)
            if not result.ok:
                self.print_error_message(result)
        else:
            self.logger.info(
                'Restart is not required. Add --force-restart if you need to restart anyway')

    def update_setting(self, key, value):
        """Update a setting"""
        self.logger.info('Updating setting "' + key + '"')
        result = requests.post(
            self.sonarqube_host + "/api/settings/set",
            auth=self.auth(),
            verify=self.verify_certificate,
            params={"key": key, "value": value})
        if not result.ok:
            self.print_error_message(result)

    def update_plugin(self, plugin):
        """Update a plugin to its latest version"""

        self.logger.info('Updating : ' + plugin + '...')
        result = requests.post(
            self.sonarqube_host + "/api/plugins/update",
            auth=self.auth(),
            verify=self.verify_certificate,
            params={"key": plugin})
        if not result.ok:
            self.print_error_message(result)

    def get_error_message(self, result):
        return "\n".join(
            [message.get('msg') for message in result.json().get('errors')])

    def print_error_message(self, result):
        self.logger.error(" Error : " + self.get_error_message(result))

    def update_plugins(self, plugins):
        """Update a list of plugins to their latest versions"""

        [self.update_plugin(plugin) for plugin in plugins]

    def update_installed_plugins(self, update_all_plugins=False):
        """Update installed plugins to their latest versions"""

        if update_all_plugins:
            plugins = [plugin.get('key') for plugin in self.plugins_installed()]
        else:
            plugins = [plugin.get('key') for plugin in self.plugins_to_update()]

        self.logger.info('Update plugins : ' + ", ".join(plugins))
        self.update_plugins(plugins)

    def has_pended_plugins(self):
        """Informs if a plugins has been installed, updated or removed waiting for server restart"""

        self.logger.info('Checking for plugins pending...')
        result = requests.post(
            self.sonarqube_host + "/api/plugins/pending",
            auth=self.auth(),
            verify=self.verify_certificate)

        if not result.ok:
            self.logger.error("  - Error : " + "\n".join(
                [message.get('msg') for message in result.json().get('errors')]))

        plugins_are_pending = (len(result.json().get('installing')) > 0
                               or len(result.json().get('updating')) > 0
                               or len(result.json().get('installing')) > 0)

        if plugins_are_pending:
            self.logger.info('Some plugins are pending')
        return plugins_are_pending

    def bootstrap_project_security_groups(self, project_info:list):
        (project_key, project_name) = project_info

        # create project if it doesn't exist
        if not self.__projects_exist([project_key]):
            project_created = self.__create_project(project_key, project_name)
            self.logger.info(f'The project {project_created} was created.')
        else:
            self.logger.info(f'The project {project_key} already exists.')

        if not self.__projects_exist([project_key]):
            raise AssertionError(
                "The project {} still does't exist. Abort.".format(project_key))

        # Here, assume the project does exist

        # create admin group
        group_name = '{}---admins'.format(project_key)
        group_description = '{} project admin group'.format(project_key)
        admin_group_created = self.__create_group(group_name, description=group_description)
        self.logger.info(f'The group {group_name} has been created.')
        self.logger.debug(admin_group_created)

        for permission in ['admin']:
            self.__add_group_permission_to_project(group_name, project_key, permission)


        # create developer group
        group_name = '{}---devs'.format(project_key)
        group_description = '{} project developers group'.format(project_key)
        dev_group_created = self.__create_group(
            group_name, description=group_description)
        self.logger.info(f'The group {group_name} has been created. Response : <{dev_group_created}>')

        for permission in ['codeviewer', 'user', 'issueadmin', 'securityhotspotadmin']:
            self.__add_group_permission_to_project(
                group_name, project_key, permission)

    def __get_user_tokens(self, user_name:str) -> list:
        res = self.__get(
            api_url="/api/user_tokens/search",
            api_params={
                'login': user_name
            }
        )

        if res.ok:
            user_tokens = res.json()
            for current_token in user_tokens['userTokens']:
                yield current_token

    def __get_user_token(self, user_name:str, token_name:str) -> dict:
        for token in self.__get_user_tokens(user_name):
            if token['name'] == token_name:
                return token_name

    def __generate_user_token(self, user_name: str, token_name: str) -> str:
        self.logger.info(f'Generating the token "{token_name} for the user {user_name}...')
        token_response = self.__post(
            api_url='/api/user_tokens/generate',
            api_params={
                'login': user_name,
                'name': token_name
            }
        )

        if token_response.ok:
            return token_response

        return {}

    def __find_user_by_field(self, value_to_look_for:str, field_to_use:str) -> list:
        if field_to_use not in ['login', 'email', 'name']:
            raise ValueError(f'The field {field_to_use} is not allowed')

        res = self.__get(
            api_url="/api/users/search",
            api_params={
                'q': value_to_look_for
            }
        )

        if res.ok:
            json_search_users = res.json()
            if json_search_users['paging']['total'] == 0:
                return []

            return [current_user for current_user in json_search_users['users'] if current_user[field_to_use] == value_to_look_for]

        return []

    def __user_login_exists(self, user_login: str) -> bool:
        return len(self.__find_user_by_field(user_login, 'login')) > 0

    def __create_user(self, login, name, password, email='', local='true'):
        res = self.__post(
            api_url="/api/users/create",
            api_params={
                'email': email,
                'local': local,
                'login': login,
                'name': name,
                'password': password,
            }
        )

        if res.ok:
            return res.json()

        return {}

    def create_azure_user(self, keyvault_name:str):
        # store the token into a keyvault
        # will use the environment variables AZURE_CLIENT_ID, AZURE_TENANT_ID and AZURE_CLIENT_SECRET
        credential = EnvironmentCredential()

        secret_client = SecretClient(
            vault_url=f"https://{keyvault_name}.vault.azure.net/",
            credential=credential
        )

        # Check if the azure user already exist
        if not self.__user_login_exists('azure'):
            self.logger.info('The azure user does not exist, create it.')

            # Retrieve the azure user password as stored in the key vault
            azure_user_password = secret_client.get_secret(
                'SonarQubeAzurePassword')

            user_created = self.__create_user('azure', 'Azure', azure_user_password.value,
                email='TEAM-cloudninjas@Faurecia.onmicrosoft.com', local='true')
            self.logger.debug(user_created)
        else:
            self.logger.info('The azure user already exists.')

        if not self.__user_login_exists('azure'):
            raise AssertionError(f"The user azure still does't exist. Abort.")

        self.logger.info('Setting the Azure user permissions')
        for permission in ['admin', 'scan', 'provisioning']:
            self.__add_global_permission_to_user('azure', permission)

        for permission in ['admin', 'scan']:
            self.__add_user_permission_to_template('azure', 'default_template', permission)

        # generate the access token if it doesn't exist
        token_parameters = ('azure', 'azure_devops_service_connection')
        if not self.__get_user_token(*token_parameters):
            generated_token = self.__generate_user_token(*token_parameters)
            if generated_token.ok:
                token_json = generated_token.json()

                secret = secret_client.set_secret(
                    "SonarQubeAzureToken", token_json['token'])

                self.logger.info(
                    f'The generated token for the user "azure" has been stored in the secret "{secret.name}" of the keyvault {keyvault_name}.')

        else:
            self.logger.info(f'The token "{token_parameters[1]}" already exists.')


    def bootstrap_security_groups(self):

        self.__create_group("dev-team-leader", "Development team leader that can manage quality profiles and gates.")
        self.__add_group_permission_to_template("dev-team-leader", "default_template", "user")
        self.__add_group_permission_to_template("dev-team-leader", "default_template", "codeviewer")
        self.__add_group_permission_to_template("dev-team-leader", "default_template", "issueadmin")
        self.__add_group_permission_to_template("dev-team-leader", "default_template", "securityhotspotadmin")
        self.__add_group_permission_to_template("dev-team-leader", "default_template", "admin")

        self.__create_group("developer", "User that need to see issues and advanced reports and source code.")
        self.__add_group_permission_to_template("developer", "default_template", "user")
        self.__add_group_permission_to_template("developer", "default_template", "codeviewer")

        self.__create_group("stakeholder",
                            "This user is generally manager or stakeholder interested in high level reports.")
        self.__add_group_permission_to_template("stakeholder", "default_template", "user")

        # reduce default user permissions to nothing
        self.__remove_group_permission_from_template("sonar-users", "default_template", "codeviewer")
        self.__remove_group_permission_from_template("sonar-users", "default_template", "user")

    def __projects_exist(self, project_names: list):
        res = requests.get(
            self.sonarqube_host + "/api/projects/search",
            auth=self.auth(),
            params={
                'projects': ','.join(project_names)
            },
            verify=self.verify_certificate
        )

        json_response = res.json()
        if 'errors' in json_response:
            self.logger.error(self.get_error_message(res))

        return [current_item['key'] for current_item in json_response['components']] if 'components' in json_response else []

    def __create_project(self, project_key:str, project_name: str, visibility='private'):
        self.logger.info('Creating project {} ...'.format(project_key))

        # The project key max length is 400 characters
        project_key = project_key[:401]

        res = requests.post(
            self.sonarqube_host + "/api/projects/create",
            auth=self.auth(),
            params={
                'project': project_key,
                'name': project_name,
                'visibility': visibility
                },
            verify=self.verify_certificate
        )

        return res.json()

    def __create_group(self, name: str, description=''):
        if self.__group_exists(name):
            self.logger.info('The group "{}" already exists.'.format(name))
            return {}

        self.logger.info('Creating group "{}" ...'.format(name))
        res = requests.post(
            self.sonarqube_host + "/api/user_groups/create",
            auth=self.auth(),
            params={
                "description": description,
                "name": name
            },
            verify=self.verify_certificate
        )

        return res.json()

    def __group_exists(self, group_name) -> bool:
        self.logger.info('Searching for group "{}"...'.format(group_name))
        res = requests.get(
            self.sonarqube_host + "/api/user_groups/search",
            auth=self.auth(),
            params={
                "q": group_name,
                "ps": 10, # page size
                "f": "name" # fields to return
            },
            verify=self.verify_certificate
        )

        json_response = res.json()
        for group in json_response['groups']:
            if group['name'] == group_name:
                return True

        return False

    def __add_group_permission_to_project(self, group_name:str, project_key:str, permission):
        self.logger.info('Adding group {} permission {} to project {} ...'.format(
            group_name, permission, project_key))
        requests.post(
            self.sonarqube_host + "/api/permissions/add_group",
            auth=self.auth(),
            params={
                "projectKey": project_key,
                "groupName": group_name,
                "permission": permission
            },
            verify=self.verify_certificate
        )

    def __add_user_permission_to_template(self, user_name, template, permission):
        self.logger.info(
            f'Adding user "{user_name}" permission "{permission}" to template "{template}"...')
        self.__post(
            api_url="/api/permissions/add_user_to_template",
            api_params={
                "templateId": template,
                "login": user_name,
                "permission": permission
            }
        )

    def __add_group_permission_to_template(self, group_name, template, permission):
        self.logger.info('Adding group "' + group_name + '" permission "' +
                         permission + '" to template "' + template + '"...')
        requests.post(
            self.sonarqube_host + "/api/permissions/add_group_to_template",
            auth=self.auth(),
            params={"templateId": template, "groupName": group_name, "permission": permission},
            verify=self.verify_certificate)

    def __remove_group_permission_from_template(self, group_name, template, permission):
        self.logger.info('Removing group "' + group_name + '" permission "' +
                         permission + '" from template "' + template + '"...')
        requests.post(
            self.sonarqube_host + "/api/permissions/remove_group_from_template",
            auth=self.auth(),
            params={"templateId": template, "groupName": group_name, "permission": permission},
            verify=self.verify_certificate)

    def __add_global_permission_to_user(self, user_name: str, permission: str):
        self.logger.info(
            f'Adding the permission {permission} to user {user_name}...')
        self.__post(
            api_url="/api/permissions/add_user",
            api_params={
                'login': user_name,
                'permission': permission
            }
        )

