#!/bin/bash
#
# This script will ensure that the service connection is using the token
# generated on SonarQube
#
# This script is not used in this Helm Chart directly, but in an Azure DevOps Pipeline

set -eu

SCRIPT_PATH="$(dirname "${BASH_SOURCE[0]}")"
DEBUG="${DEBUG:-false}"

if [[ $DEBUG == "true" ]]; then
    set -x
fi

AZ_TENANT_ID="${AZ_TENANT_ID:?Not set in environment}"
AZ_CLIENT_SECRET="${AZ_CLIENT_SECRET:?Not set in environment}"
AZ_CLIENT_ID="${AZ_CLIENT_ID:?Not set in environment}"
AZ_DEVOPS_ORG="${AZ_DEVOPS_ORG:?Not set in environment}"
AZ_DEVOPS_PROJECT="${AZ_DEVOPS_PROJECT:?Not set in environment}"
AZ_APP_KV_NAME="${AZ_APP_KV_NAME:?Not set in environment}"
AZ_APP_KV_SECRET_NAME="${AZ_APP_KV_SECRET_NAME:?Not set in environment}"

SNR_SERVICE_CONNECTION="${SNR_SERVICE_CONNECTION:?Not set in environment}"
ADMIN_PASSWORD="${ADMIN_PASSWORD:?Not set in environment}"
SERVER_BASE_URL="${SERVER_BASE_URL:?Not set in environment}"

function _az_login_check() {
    printf "* Check connectivity with Azure...\\n"
    if az account list >/dev/null
    then
        printf "  ✔ We are connected to Azure.\\n"
        return 0
    else
        printf "  ❌ Cannot connect on Azure.\\n"
        exit 2
    fi
}

# Ensure az cli can be invoke
_az_login_check

# First, install the azure-devops extension without prompting for confirmation
extension_to_install="azure-devops"
printf "* Installing the az extension %s...\\n" "${extension_to_install}"
az extension add --name "${extension_to_install}" --yes

printf "* Updating the az extension %s...\\n" "${extension_to_install}"
az extension update --name "${extension_to_install}" |:

printf "* Outputting the az extension information...\\n"
az extension show --name "${extension_to_install}" --query '{name:name,version:version, path:path}' --output tsv

printf "* Install python dependencies...\\n"
/usr/bin/env python3 --version
/usr/bin/env python3 -m pip install --no-cache-dir --quiet --upgrade --user virtualenv
/usr/bin/env python3 -m venv .venv
source .venv/bin/activate
/usr/bin/env python3 -m pip install --no-cache-dir --quiet --upgrade pip setuptools wheel
/usr/bin/env python3 -m pip install --no-cache-dir --quiet --upgrade --requirement "${SCRIPT_PATH}/../python/requirements.txt"
printf "* Generate admin Token...\\n"
AZURE_CLIENT_ID="${AZ_CLIENT_ID}" AZURE_CLIENT_SECRET="${AZ_CLIENT_SECRET}" AZURE_TENANT_ID="${AZ_TENANT_ID}" /usr/bin/env python3 "${SCRIPT_PATH}/../python/configure.py" "${SERVER_BASE_URL}" "admin" --auth-type basic --password "${ADMIN_PASSWORD}" --create-azure-user "${AZ_APP_KV_NAME}"

printf "* Get %s/%s secret...\\n" "${AZ_APP_KV_NAME}" "${AZ_APP_KV_SECRET_NAME}"
MY_LITTLE_TOKEN=$(az keyvault secret show \
    --vault-name "${AZ_APP_KV_NAME}" \
    --name "${AZ_APP_KV_SECRET_NAME}" \
    --query value \
    --output tsv)

# Modify service connection
TEMP_JSON=$(mktemp)
# TEMP_JSON=$(mktemp)
UPDATED_JSON=$(mktemp)

printf "* Get %s service-endpoint id...\\n" "${SNR_SERVICE_CONNECTION}"
az devops service-endpoint list \
  --org "${AZ_DEVOPS_ORG}" \
  --project "${AZ_DEVOPS_PROJECT}" \
  --query "[?name == '${SNR_SERVICE_CONNECTION}']|[0]" \
  --output json > "${TEMP_JSON}"

endpointId=$(jq '.id' --raw-output "${TEMP_JSON}")
jq ".authorization.parameters.username = \"${MY_LITTLE_TOKEN}\"" "${TEMP_JSON}" > "${UPDATED_JSON}"
rm "${TEMP_JSON}"

printf "* Update %s service-endpoint token...\\n" "${SNR_SERVICE_CONNECTION}"
az devops invoke \
  --org "${AZ_DEVOPS_ORG}" \
  --area serviceendpoint \
  --resource endpoints \
  --http-method "PUT" \
  --api-version "5.1-preview" \
  --route-parameters project="${AZ_DEVOPS_PROJECT}" endpointId="${endpointId}" \
  --in-file "${UPDATED_JSON}" \
  --output none

rm "${UPDATED_JSON}"

printf "  ✔ All done.\\n"
